from __future__ import annotations

from typing import Any
from typing import Optional

import optuna


SYSTEM_ATTR_NAME = "dashboard:objective_names"


def set_objective_names(study: optuna.Study, names: list[str]) -> None:
    """Set the names of objectives.

    Example:

       .. code-block:: python

          import optuna
          from optuna_dashboard import set_objective_names

          study = optuna.create_study(directions=["minimize", "minimize"])
          set_objective_names(study, ["val_loss", "flops"])
    """
    storage = study._storage
    study_id = study._study_id

    directions = storage.get_study_directions(study_id)
    if len(directions) != len(names):
        raise ValueError("names must be the same length with the number of objectives.")
    storage.set_study_system_attr(study_id, SYSTEM_ATTR_NAME, names)


def get_objective_names(system_attrs: dict[str, Any]) -> Optional[list[str]]:
    return system_attrs.get(SYSTEM_ATTR_NAME)



from typing import BinaryIO
from typing import Protocol


class ArtifactBackend(Protocol):
    def open(self, artifact_id: str) -> BinaryIO:
        ...

    def write(self, artifact_id: str, content_body: BinaryIO) -> None:
        ...

    def remove(self, artifact_id: str) -> None:
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

import boto3


if TYPE_CHECKING:
    from typing import BinaryIO
    from typing import Optional

    from mypy_boto3_s3 import S3Client


class Boto3Backend:
    """An artifact backend for S3.

    Example:
       .. code-block:: python

          import optuna
          from optuna_dashboard.artifact import upload_artifact
          from optuna_dashboard.artifact.boto3 import Boto3Backend

          artifact_backend = Boto3Backend("my-bucket")

          def objective(trial: optuna.Trial) -> float:
              ... = trial.suggest_float("x", -10, 10)
              file_path = generate_example_png(...)
              upload_artifact(artifact_backend, trial, file_path)
              return ...
    """

    def __init__(self, bucket_name: str, client: Optional[S3Client] = None) -> None:
        self.bucket = bucket_name
        self.client = client or boto3.client("s3")

    def open(self, artifact_id: str) -> BinaryIO:
        obj = self.client.get_object(Bucket=self.bucket, Key=artifact_id)
        body = obj.get("Body")
        assert body is not None
        return body  # type: ignore

    def write(self, artifact_id: str, content_body: BinaryIO) -> None:
        self.client.upload_fileobj(content_body, self.bucket, artifact_id)

    def remove(self, artifact_id: str) -> None:
        self.client.delete_object(Bucket=self.bucket, Key=artifact_id)


if TYPE_CHECKING:
    # A mypy-runtime assertion to ensure that Boto3Backend
    # implements all abstract methods in ArtifactBackendProtocol.
    from .protocol import ArtifactBackend

    _: ArtifactBackend = Boto3Backend("")



from __future__ import annotations

from dataclasses import dataclass
import json
from typing import TYPE_CHECKING
from typing import Union

import optuna


if TYPE_CHECKING:
    from typing import Any
    from typing import Literal
    from typing import Optional
    from typing import TypedDict

    ChoiceWidgetJSON = TypedDict(
        "ChoiceWidgetJSON",
        {
            "type": Literal["choice"],
            "description": Optional[str],
            "choices": list[str],
            "values": list[float],
        },
    )
    SliderWidgetLabel = TypedDict(
        "SliderWidgetLabel",
        {"value": float, "label": str},
    )
    SliderWidgetJSON = TypedDict(
        "SliderWidgetJSON",
        {
            "type": Literal["slider"],
            "description": Optional[str],
            "min": float,
            "max": float,
            "step": Optional[float],
            "labels": Optional[list[SliderWidgetLabel]],
        },
    )
    TextInputWidgetJSON = TypedDict(
        "TextInputWidgetJSON",
        {"type": Literal["text"], "description": Optional[str]},
    )
    UserAttrRefJSON = TypedDict("UserAttrRefJSON", {"type": Literal["user_attr"], "key": str})
    ObjectiveFormWidgetJSON = Union[
        ChoiceWidgetJSON, SliderWidgetJSON, TextInputWidgetJSON, UserAttrRefJSON
    ]


@dataclass
class ObjectiveChoiceWidget:
    choices: list[str]
    values: list[float]
    description: Optional[str] = None

    def to_dict(self) -> ChoiceWidgetJSON:
        return {
            "type": "choice",
            "description": self.description,
            "choices": self.choices,
            "values": self.values,
        }


@dataclass
class ObjectiveSliderWidget:
    min: float
    max: float
    step: Optional[float] = None
    labels: Optional[list[tuple[float, str]]] = None
    description: Optional[str] = None

    def to_dict(self) -> SliderWidgetJSON:
        labels: Optional[list[SliderWidgetLabel]] = None
        if self.labels is not None:
            labels = [{"value": value, "label": label} for value, label in self.labels]
        return {
            "type": "slider",
            "description": self.description,
            "min": self.min,
            "max": self.max,
            "step": self.step,
            "labels": labels,
        }


@dataclass
class ObjectiveTextInputWidget:
    description: Optional[str] = None

    def to_dict(self) -> TextInputWidgetJSON:
        return {
            "type": "text",
            "description": self.description,
        }


@dataclass
class ObjectiveUserAttrRef:
    key: str

    def to_dict(self) -> UserAttrRefJSON:
        return {
            "type": "user_attr",
            "key": self.key,
        }


ObjectiveFormWidget = Union[
    ObjectiveChoiceWidget, ObjectiveSliderWidget, ObjectiveTextInputWidget, ObjectiveUserAttrRef
]
SYSTEM_ATTR_KEY = "dashboard:objective_form_widgets:v1"


def register_objective_form_widgets(
    study: optuna.Study, widgets: list[ObjectiveFormWidget]
) -> None:
    if len(study.directions) != len(widgets):
        raise ValueError("The length of actions must be the same with the number of objectives.")
    widget_dicts = [w.to_dict() for w in widgets]
    study._storage.set_study_system_attr(study._study_id, SYSTEM_ATTR_KEY, widget_dicts)


def get_objective_form_widgets_json(
    study_system_attr: dict[str, Any]
) -> Optional[list[ObjectiveFormWidgetJSON]]:
    if SYSTEM_ATTR_KEY in study_system_attr:
        return study_system_attr[SYSTEM_ATTR_KEY]
    # For optuna-dashboard v0.9.0b5 users
    if "dashboard:objective_form_widgets" in study_system_attr:
        return json.loads(study_system_attr["dashboard:objective_form_widgets"])
    return None



from ._app import run_server  # noqa
from ._app import wsgi  # noqa
from ._named_objectives import set_objective_names  # noqa
from ._note import save_note  # noqa
from ._objective_form_widget import ObjectiveChoiceWidget  # noqa
from ._objective_form_widget import ObjectiveSliderWidget  # noqa
from ._objective_form_widget import ObjectiveTextInputWidget  # noqa
from ._objective_form_widget import ObjectiveUserAttrRef  # noqa
from ._objective_form_widget import register_objective_form_widgets  # noqa


__version__ = "0.9.0"
